SystemOrganization addCategory: #'CTranslator-Hierarchy'!
SystemOrganization addCategory: #'CTranslator-Translators'!
SystemOrganization addCategory: #'CTranslator-Exceptions'!
SystemOrganization addCategory: #'CTranslator-Programs'!
SystemOrganization addCategory: #'CTranslator-Output'!

RBProgramNodeVisitor subclass: #CInliner
	instanceVariableNames: 'class methodNode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CInliner class methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 17:58'!
inline: selector of: aClass
	^ self new inline: selector of: aClass! !

!CInliner methodsFor: 'visiting' stamp: 'Richo 7/22/2011 19:44'!
acceptMessageNode: aMessageNode 
	| method receiverClass |
	receiverClass := CTypeCollector findTypeOf: aMessageNode receiver in: class.
	method := receiverClass ultimateCompiledMethodAt: aMessageNode selector ifAbsent: [self error: 'Method not found'].
	

	((method pragmas anySatisfy: [:each | each keyword = #c:]) not and: [
	(method pragmas anySatisfy: [:each | each keyword = #inlined])
		or: [receiverClass isMeta]])
		ifTrue: [self replace: aMessageNode with: (self class inline: method selector of: receiverClass)]
		ifFalse: [super acceptMessageNode: aMessageNode]! !

!CInliner methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 19:34'!
addTempVarNamed: varName ofType: varType
	| pragma |
	"Check if the variable is already added"
	(self currentMethod body temporaries anySatisfy: [:each | each name = varName])
		ifTrue: [^ self].
	"Add variable to temporaries"
	self currentMethod body temporaries: self currentMethod body temporaries, {self newVariableNamed: varName}.
	"Add type to #tempTypes: pragma"
	pragma := self currentMethod pragmas 
		detect: [:each | each selector = #tempTypes:]
		ifNone: [self createTempTypesPragma].
	pragma arguments:{RBLiteralArrayNode value: pragma arguments first value, {varType name}} asOrderedCollection
	
! !

!CInliner methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 19:18'!
appendTempVarsFrom: aMethodNode
	aMethodNode body temporaries do: [:temp || varName varType var |
		varName := self varNameFor: temp in: aMethodNode.
		varType := CTypeCollector findTypeOf: temp in: class.
		
		"Add var to currentMethod temporaries and #tempTypes: pragma"
		self addTempVarNamed: varName ofType: varType]! !

!CInliner methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 19:03'!
createTempTypesPragma
	| pragma |
	pragma := RBPragmaNode selectorParts: {RBKeywordToken value: 'tempTypes:' start: nil} arguments: (OrderedCollection with: (RBLiteralArrayNode value: #())).
	self currentMethod pragmas: (self currentMethod pragmas, {pragma}) asOrderedCollection.
	^ pragma! !

!CInliner methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 19:16'!
createTempVarsForArgumentsOf: aMethodNode
	aMethodNode arguments do: [:arg || varName varType var |
		varName := self varNameFor: arg in: aMethodNode.
		varType := CTypeCollector findTypeOf: arg in: class.
		
		"Add var to currentMethod temporaries and #tempTypes: pragma"
		self addTempVarNamed: varName ofType: varType]! !

!CInliner methodsFor: 'accessing' stamp: 'Richo 7/22/2011 18:00'!
currentMethod
	^ methodNode! !

!CInliner methodsFor: 'inlining' stamp: 'Richo 7/22/2011 17:59'!
inline: selector of: aClass
	class := aClass.
	methodNode := RBParser parseMethod: (class ultimateSourceCodeAt: selector ifAbsent: [self error: 'Source code not found']).
	self visitNode: methodNode.
	^ methodNode! !

!CInliner methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 19:31'!
newAssignment: aVariableNode value: aValueNode
	^ RBAssignmentNode variable: aVariableNode value: aValueNode! !

!CInliner methodsFor: 'accessing' stamp: 'Richo 7/22/2011 18:09'!
newVariableNamed: aByteString 
	^ RBVariableNode named: aByteString! !

!CInliner methodsFor: 'inlining' stamp: 'Richo 7/22/2011 19:30'!
replace: aMessageNode with: methodNode
	| sequence statement |
	statement := aMessageNode statementNode.
	sequence := statement parent.
	
	"Create temp vars for methodNode's args and add them to currentMethod temporaries and #tempTypes: pragma"
	self createTempVarsForArgumentsOf: methodNode.
	
	"Add methodNode temps and #tempTypes: to currentMethod"
	self appendTempVarsFrom: methodNode.
	
	"Store assignments to the tempVars representing arguments"
	methodNode arguments with: aMessageNode arguments do: [:arg :value || temp |
		temp := self newVariableNamed: (self varNameFor: arg in: methodNode).
		sequence addNode: (RBAssignmentNode variable: temp value: value)
			before: statement].

	"Replace methodNode arg and temp references with tempVars references"
	methodNode arguments, methodNode body temporaries do: [:arg || temp |
		temp := self newVariableNamed: (self varNameFor: arg in: methodNode).
		methodNode nodesDo: [:node | node = arg ifTrue: [node replaceWith: temp copy]]].
	
	"Store methodNode's body without return"
	methodNode nodesDo: [:node | node isReturn ifTrue: [node replaceWith: node value]].
	methodNode body statements do: [:each | sequence addNode: each before: statement].

	aMessageNode isUsed
		ifTrue: ["If aMessageNode is used create a temp var to hold methodNode's result"
			| temp |
			temp := self newVariableNamed: methodNode selector asCIdentifier, '_return_'.
			self addTempVarNamed: temp name ofType: (CTypeCollector findTypeOf: methodNode in: class).
			methodNode body statements last
				replaceWith: (self newAssignment: temp value: methodNode body statements last copy).
			aMessageNode replaceWith: temp copy.
			]
		ifFalse: ["If aMessageNode is not used remove statement"
			sequence removeNode: statement]	
! !

!CInliner methodsFor: 'accessing' stamp: 'Richo 7/22/2011 18:09'!
varNameFor: aVariableNode in: aMethodNode
	^ aMethodNode selector asCIdentifier, aVariableNode name asCIdentifier 
	! !

RBProgramNodeVisitor subclass: #CInlinerViejo
	instanceVariableNames: 'class'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CInlinerViejo class methodsFor: 'as yet unclassified' stamp: 'Richo 7/21/2011 16:00'!
inline: selector of: aClass
	^ self new inline: selector of: aClass! !

!CInlinerViejo methodsFor: 'visiting' stamp: 'Richo 7/22/2011 17:19'!
acceptMessageNode: aMessageNode 
	| method receiverClass |
	receiverClass := CTypeCollector findTypeOf: aMessageNode receiver in: class.
	method := receiverClass ultimateCompiledMethodAt: aMessageNode selector ifAbsent: [self error: 'Method not found'].
	
	((method pragmas anySatisfy: [:each | each keyword = #inlined])
		or: [receiverClass isMeta])
		ifTrue: [self replace: aMessageNode with: (self class inline: method selector of: receiverClass)]
		ifFalse: [super acceptMessageNode: aMessageNode]! !

!CInlinerViejo methodsFor: 'inlining' stamp: 'Richo 7/22/2011 16:47'!
addTemporaries: variables to: methodNode
	self halt.
! !

!CInlinerViejo methodsFor: 'inlining' stamp: 'Richo 7/22/2011 16:47'!
addTemporariesViejo: variablesDict to: methodNode
	| pragma |
	pragma := (methodNode pragmas
		detect: [:each | each selector = #tempTypes:]
		ifNone: [methodNode pragmas
			add: (RBPragmaNode selectorParts: {RBKeywordToken value: 'tempTypes:' start: nil} arguments: (OrderedCollection with: (RBLiteralArrayNode value: #())))]).
	variablesDict keysAndValuesDo: [:key :value |
		(methodNode body temporaries includes: value)
			ifFalse: [| tempTypes |
				methodNode body temporaries isArray ifTrue: [methodNode body temporaries: OrderedCollection new].
				methodNode body temporaries add: value.
				tempTypes := pragma arguments first value asOrderedCollection 
					add: (CTypeCollector findTypeOf: variablesDict keys first in: class) name;
					yourself.
				pragma arguments: {RBLiteralArrayNode value: tempTypes asArray} asOrderedCollection]].
	! !

!CInlinerViejo methodsFor: 'inlining' stamp: 'Richo 7/21/2011 16:38'!
inline: selector of: aClass
	| methodNode |
	class := aClass.
	methodNode := RBParser parseMethod: (class ultimateSourceCodeAt: selector ifAbsent: [self error: 'Source code not found']).
	self visitNode: methodNode.
	^ methodNode! !

!CInlinerViejo methodsFor: 'inlining' stamp: 'Richo 7/22/2011 16:52'!
replace: aMessageNode with: methodNode
	"It replaces the given message node with the method's implementation. It's not as simple as it sounds, though, because it makes it work even when the method has parameters. To do that, it creates temporary variables to hold the values of the method's arguments and it replaces the references to arguments with this newly created temp vars"
	| statement sequence tempVars |
	statement := aMessageNode statementNode.
	sequence := statement parent.
	tempVars := self tempVarsFor: methodNode.
	self addTemporaries: tempVars to: aMessageNode methodNode;
		replaceArgumentsOf: methodNode with: tempVars;
		storeAssignmentsTo: tempVars 
		values: aMessageNode arguments 
		in: sequence
		before: statement.
	aMessageNode = statement 
		ifTrue: [	methodNode body statements do: [:each | sequence addNode: each before: statement].
				sequence removeNode: statement]
		ifFalse: [| returnVar |
				self halt.
				methodNode body statements do: [:each | sequence addNode: each before: statement].
				returnVar := (RBVariableNode named: methodNode selector asCIdentifier , '_return_').
				methodNode body statements last 
					replaceWith: (RBAssignmentNode variable: returnVar value: methodNode body statements last copy).
				self addTemporaries: (Dictionary new at: methodNode put: returnVar; yourself) to: aMessageNode methodNode.
				aMessageNode replaceWith: returnVar.

			]! !

!CInlinerViejo methodsFor: 'inlining' stamp: 'Richo 7/21/2011 15:41'!
replaceArgumentsOf: parseNode with: variablesDict 
	(variablesDict includesKey: parseNode)
		ifTrue: [parseNode replaceWith: (variablesDict at: parseNode) copy]
		ifFalse: [parseNode children 	do: [:each | self replaceArgumentsOf: each with: variablesDict]]! !

!CInlinerViejo methodsFor: 'inlining' stamp: 'Richo 7/21/2011 17:16'!
storeAssignmentsTo: variablesDict values: variablesValues in: sequence before: aNode
	variablesDict values with: variablesValues do: [:var :value |
		sequence addNode: (RBAssignmentNode variable: var value: value)
			before: aNode]
! !

!CInlinerViejo methodsFor: 'inlining' stamp: 'Richo 7/22/2011 16:46'!
tempVarsFor: methodNode 
	| variables |
	variables := OrderedCollection new.
	methodNode arguments do: [:arg || varName |
		varName := methodNode selector asCIdentifier, arg name asCIdentifier.
		variables add: (RBVariableNode named: varName) -> (CTypeCollector findTypeOf: arg in: class)].
	methodNode body temporaries do: [:var |
		variables add: var -> (CTypeCollector findTypeOf: var in: class)].
	^ variables! !

RBProgramNodeVisitor subclass: #CMethodTranslator
	instanceVariableNames: 'source class builder'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CMethodTranslator class methodsFor: 'instance creation' stamp: 'Richo 6/28/2011 11:29'!
new
	^ self shouldNotImplement! !

!CMethodTranslator class methodsFor: 'translating' stamp: 'Richo 6/28/2011 11:28'!
translate: selector of: class with: aBuilder
	^ (self withBuilder: aBuilder) translate: selector of: class! !

!CMethodTranslator class methodsFor: 'instance creation' stamp: 'Richo 6/28/2011 11:29'!
withBuilder: aBuilder
	^ self basicNew initializeWith: aBuilder! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/30/2011 12:43'!
acceptAssignmentNode: anAssignmentNode
	self visitNode: anAssignmentNode variable.
	source nextPutAll: ' = '.
	self visitNode: anAssignmentNode value.
	source nextPut: $;; cr.! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptCascadeNode: aCascadeNode
	aCascadeNode messages do: [:each | self visitNode: each. source nextPut: $;; cr].! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptLiteralNode: aLiteralNode
	| value |
	value := aLiteralNode value.
	value isFloat 
		ifTrue: [source nextPutAll: (value printShowingDecimalPlaces: 12).
			^ self].
	value isString 
		ifTrue: [source nextPut: $"; nextPutAll: value asString; nextPut: $".
			^ self].
	source nextPutAll: value asString! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 7/26/2011 15:32'!
acceptMessageNode: aMessageNode 
	| method receiverClass |
	receiverClass := CTypeCollector findTypeOf: aMessageNode receiver in: class.
	method := receiverClass ultimateCompiledMethodAt: aMessageNode selector ifAbsent: [self error: 'Method not found'].

	"Is it a c primitive?"
	(method pragmaAt: #c:)
		ifNotNil: [:nxcPrim | source nextPutAll: ((CPrimitiveRule on: nxcPrim arguments first) applyTo: aMessageNode using: self).
			^ self].

	"Is it a creation method?"
	(method pragmaAt: #creationMethod)
		ifNotNil: [:pragma | self storeInstanceCreation: receiverClass.
			^ self].

	"If we got here, it is a regular method call and we need to signal the MethodToImplement exception."
	(CMethodToImplement
		class: receiverClass
		selector: aMessageNode selector)
			signal.
	self storeRegularMessageSend: aMessageNode! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 7/27/2011 17:59'!
acceptMethodNode: aMethodNode
	"I make sure the last statement is a return self"
	super acceptMethodNode: aMethodNode.
	aMethodNode lastIsReturn ifFalse: [self visitNode: self selfReturn]! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptPragmaNode: aPragmaNode
	"Do nothing"! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptReturnNode: aReturnNode
	source nextPutAll: 'return '.
	self visitNode: aReturnNode value.
	source nextPut: $;; cr.! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 7/27/2011 17:58'!
acceptSequenceNode: aSequenceNode
aSequenceNode statements do: [:each | 
	self visitNode: each.
	each isMessage ifTrue: [source nextPut: $;; cr]]! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 7/26/2011 15:28'!
acceptVariableNode: aVariableNode
	"Is it a non primitive value outside of variable assignment?"
	"((CTypeCollector findTypeOf: aVariableNode in: class) isPrimitive not 
		and: [(aVariableNode parent isAssignment) not])
			ifTrue: [source nextPut: $&]."
			
	"Is it an inst var?"
	(class instVarNames includes: aVariableNode name)
		ifTrue: [(CInstVarToImplement class: class var: aVariableNode) signal.
			source nextPutAll: 'self->'].		

	"Is it 'self' on the class side?"
	(class isMeta and: [aVariableNode name = 'self'])
		ifTrue: [source nextPutAll: 'The_', class instanceSide asCType, '_class'.
			^ self].
	
	"Is it a class?"
	(Smalltalk includesKey: aVariableNode name asSymbol)
		ifTrue: [source nextPutAll: 'The_', aVariableNode name, '_class'.
			^ self].
	
	source nextPutAll: aVariableNode name.! !

!CMethodTranslator methodsFor: 'accessing' stamp: 'Richo 7/21/2011 16:43'!
argVarsOf: methodNode
	^ Array streamContents: [:arguments |
		arguments nextPut: (builder newVariableNamed: 'self' type: class).
		methodNode arguments do: [:each | 
			arguments nextPut: (builder
					newVariableNamed: each name
					type: (CTypeCollector findTypeOf: each in: class))]]! !

!CMethodTranslator methodsFor: 'accessing' stamp: 'Richo 6/28/2011 11:30'!
builder
	^ builder! !

!CMethodTranslator methodsFor: 'initializing' stamp: 'Richo 6/28/2011 11:34'!
initializeWith: aBuilder
	builder := aBuilder.
	self initialize! !

!CMethodTranslator methodsFor: 'accessing' stamp: 'Richo 7/27/2011 13:40'!
selfReturn
	^ RBReturnNode value: (RBVariableNode named: 'self')! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 7/26/2011 15:34'!
storeInstanceCreation: aMetaclass 
	source nextPutAll: (('({1}*)malloc(sizeof({1}))') format: {aMetaclass instanceSide name})! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/21/2011 18:00'!
storeRegularMessageSend: aMessageNode
	source nextPutAll: aMessageNode selector asCIdentifier, '('.
	{aMessageNode receiver}, aMessageNode arguments
		do: [:each | self visitNode: each]
		separatedBy: [source nextPutAll: ', '].
	source nextPut: $)! !

!CMethodTranslator methodsFor: 'accessing' stamp: 'Richo 7/21/2011 16:43'!
tempVarsOf: methodNode 
	^ Array streamContents: [:temporaries |
		methodNode allTemporaryVariables
			do: [:varName | 
				temporaries nextPut: (builder
					newVariableNamed: varName
					type: (CTypeCollector
							findTypeOf: ((RBVariableNode named: varName) parent: methodNode)
							in: class))]]! !

!CMethodTranslator methodsFor: 'translating' stamp: 'Richo 7/25/2011 15:34'!
translate: selector of: aClass
	| methodNode |
	source := WriteStream on: ''.
	class := aClass.
	methodNode := (RBParser parseMethod: (class ultimateSourceCodeAt: selector ifAbsent: [self error: 'Source code not found'])).
	self visitNode: methodNode.
	^ self builder 
		newFunctionFor: class
		node: methodNode
		body: source contents
		arguments: (self argVarsOf: methodNode)
		temporaries: (self tempVarsOf: methodNode)! !

!CMethodTranslator methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
visitArguments: aNodeCollection 
	"Do nothing"! !

!CMethodTranslator methodsFor: 'translating' stamp: 'Richo 6/20/2011 20:59'!
withClass: aClass do: aBlock
	"Replaces the class just for running the block, then it goes back to the old class.
	Very useful for translating inlined methods, which needs to use the same source but change the current class"
	| old |
	old := class.
	class := aClass.
	aBlock value.
	class := old! !

!CMethodTranslator methodsFor: 'translating' stamp: 'Richo 6/20/2011 21:12'!
withSource: aWriteStream do: aBlock
	"Replaces the source stream just for running the block, then it goes back to the old source.
	Very useful for CPrimitiveRule, which needs to translate just specific expressions without messing with already translated stuff"
	| oldSource |
	oldSource := source.
	source := aWriteStream.
	aBlock value.
	source := oldSource! !

CMethodTranslator subclass: #CppMethodTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CppMethodTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 7/27/2011 15:27'!
acceptVariableNode: aVariableNode
	(class bindingOf: aVariableNode name)
		ifNotNil: [:association |
			association value isLiteral
				ifTrue: [source nextPutAll: association value asString; space;
							nextPutAll: '/* ', association key, ' */'.
					^ self]].
	(class instVarNames includes: aVariableNode name)
		ifTrue: [(CInstVarToImplement class: class var: aVariableNode) signal].
		
	(aVariableNode name = 'self')
		ifTrue: [source nextPutAll: 'this']
		ifFalse: [source nextPutAll: aVariableNode name]! !

!CppMethodTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 7/27/2011 15:25'!
argVarsOf: methodNode
	^ Array streamContents: [:arguments |
		methodNode arguments do: [:each | 
			arguments nextPut: (builder
					newVariableNamed: each name
					type: (CTypeCollector findTypeOf: each in: class))]]! !

!CppMethodTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 7/27/2011 16:17'!
storeInstanceCreation: aMetaclass 
	source nextPutAll: (('(new {1}())') format: {aMetaclass instanceSide name})! !

!CppMethodTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 7/27/2011 16:13'!
storeRegularMessageSend: aMessageNode
	self visitNode: aMessageNode receiver.
	(CTypeCollector findTypeOf: aMessageNode receiver in: class) isMeta
		ifTrue: ["The method is static (i.e. class side)"
			source nextPutAll: '::']
		ifFalse: [source nextPutAll: '->'].
	source nextPutAll: aMessageNode selector asCIdentifier, '('.
	aMessageNode arguments
		do: [:each | self visitNode: each]
		separatedBy: [source nextPutAll: ', '].
	source nextPut: $)! !

RBProgramNodeVisitor subclass: #CTypeCollector
	instanceVariableNames: 'selfClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CTypeCollector class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 20:59'!
findTypeOf: aProgramNode in: aClass 
	^ self new findTypeOf: aProgramNode in: aClass ! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptArrayNode: anArrayNode 
	self halt! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 21:21'!
acceptBlockNode: aBlockNode 
	^ CBlock! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptCascadeNode: aCascadeNode 
	^ self visitNode: aCascadeNode messages last! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptLiteralArrayNode: aRBLiteralArrayNode 
	self halt! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 21:22'!
acceptLiteralNode: aLiteralNode
	aLiteralNode value isFloat
		ifTrue: [^ CFloat].
	aLiteralNode value isNumber
		ifTrue: [^CLong].
	(#(true false) includes: aLiteralNode value)
		ifTrue: [^ CBoolean].
	aLiteralNode value isString
		ifTrue: [^ CString].
	self halt! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 7/1/2011 16:43'!
acceptMessageNode: aMessageNode 
	"It looks for the pragma #returnType: in the method being called. If not found, it returns the receiver's class."
	| receiverClass method |
	receiverClass := self findTypeOf: aMessageNode receiver.
	method := receiverClass ultimateCompiledMethodAt: aMessageNode selector ifAbsent: [self error: 'Method not found'].
	
	(method pragmaAt: #returnType:)
		ifNotNilDo: [:pragma |
			^ (self computeTypes: pragma arguments first) first].

	^ receiverClass! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 7/27/2011 13:43'!
acceptMethodNode: aMethodNode
	"It looks for the #returnType: pragma to get the return type of the method"
	(aMethodNode pragmas detect: [:each | each selector = #returnType:] ifNone: [nil])
		ifNotNilDo: [:pragma | ^ (self computeTypes: pragma arguments first value) first].
	"^ CVoid"
	^ selfClass! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptPatternBlockNode: aRBPatternBlockNode 
	self halt! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptPatternWrapperBlockNode: aRBPatternWrapperBlockNode 
	self halt! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 6/20/2011 20:59'!
acceptSequenceNode: aSequenceNode 
	^ self visitNode: aSequenceNode statements last! !

!CTypeCollector methodsFor: 'visiting' stamp: 'Richo 7/14/2011 17:22'!
acceptVariableNode: aVariableNode 
	| methodNode findPragma |
	methodNode := aVariableNode methodNode.
	
	findPragma := [:selector | "Looks in the pragmas for the given selector and returns the arguments value"
		(methodNode pragmas detect: [:each | each selector = selector]
			ifNone: [self error: 'Missing ', selector printString, ' pragma for variable ', aVariableNode name printString])
			ifNotNilDo: [:pragma | pragma arguments first value]].

	"Is it an argument?"
	(methodNode arguments includes: aVariableNode)
		ifTrue: [
			^ (self computeTypes: (findPragma value: #argTypes:))
				at: (methodNode arguments indexOf: aVariableNode)].
		
	"Is it a temp var?"
	(methodNode allTemporaryVariables includes: aVariableNode name)
		ifTrue: [
			^ (self computeTypes: (findPragma value: #tempTypes:))
				at: (methodNode allTemporaryVariables indexOf: aVariableNode name)].
			
	"Is it an inst var?"
	(selfClass instVarNames includes: aVariableNode name)
		ifTrue: [| className |
			className := selfClass instanceVariableTypes at: (selfClass instVarNames indexOf: aVariableNode name).
		"In theory, this should never throw an error"
		^ Smalltalk
			at: className
			ifAbsent: [self error: 'Invalid type: ', className]].
	
	"Is it a class?"
	(Smalltalk includesKey: aVariableNode name asSymbol)
		ifTrue: [^ (Smalltalk at: aVariableNode name asSymbol) class].
		
	self error: 'Variable''s type not found'! !

!CTypeCollector methodsFor: 'utility' stamp: 'Richo 7/8/2011 16:08'!
computeTypes: anArray 
	"It reads the array from method tags and returns an array of c types"
	^ Array streamContents: [:outputStream | | inputStream |
	inputStream := anArray readStream.
	[inputStream atEnd]
		whileFalse: [| next |
			next := inputStream next.
			next := Smalltalk at: next ifAbsent: [
				"If the class doesn't match, it must be a message to selfClass"
				(selfClass respondsTo: next) 
					ifTrue: [selfClass perform: next]
					ifFalse: [^ self error: 'Invalid type']].
			inputStream peek = #*
				ifTrue: [inputStream next.
					next := CPointer on: next].
			outputStream nextPut: next]].
		! !

!CTypeCollector methodsFor: 'finding types' stamp: 'Richo 7/1/2011 16:52'!
findTypeOf: aProgramNode
	aProgramNode = (RBVariableNode named: 'self')
		ifTrue: [^ selfClass].
	^ self visitNode: aProgramNode.
! !

!CTypeCollector methodsFor: 'finding types' stamp: 'Richo 6/20/2011 20:59'!
findTypeOf: aProgramNode in: aClass 
	selfClass := aClass.
	^ self findTypeOf: aProgramNode! !

Exception subclass: #CInstVarToImplement
	instanceVariableNames: 'ownerClass variable'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Exceptions'!

!CInstVarToImplement class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 21:05'!
class: aClass var: aVariableNode 
	^ (self basicNew setClass: aClass var: aVariableNode) initialize.! !

!CInstVarToImplement methodsFor: 'exceptionDescription' stamp: 'Richo 6/20/2011 21:05'!
defaultAction
	"No one has handled this error, but now give them a chance to decide how to debug it.  If none handle this either then open debugger (see UnhandedError-defaultAction)"
	UnhandledError signalForException: self! !

!CInstVarToImplement methodsFor: 'accessing' stamp: 'Richo 6/20/2011 21:05'!
ownerClass
	^ ownerClass! !

!CInstVarToImplement methodsFor: 'private' stamp: 'Richo 6/20/2011 21:05'!
setClass: aClass var: aVariableNode
	ownerClass := aClass.
	variable := aVariableNode! !

!CInstVarToImplement methodsFor: 'accessing' stamp: 'Richo 6/20/2011 21:05'!
varName
	^ variable name! !

!CInstVarToImplement methodsFor: 'accessing' stamp: 'Richo 6/21/2011 15:03'!
varType
	^ CTypeCollector findTypeOf: variable in: ownerClass.! !

Exception subclass: #CMethodToImplement
	instanceVariableNames: 'methodClass methodSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Exceptions'!

!CMethodToImplement class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 21:04'!
class: aClass selector: aByteSymbol 
	^ (self basicNew setClass: aClass selector: aByteSymbol) initialize! !

!CMethodToImplement methodsFor: 'exceptionDescription' stamp: 'Richo 6/20/2011 21:04'!
defaultAction
	"No one has handled this error, but now give them a chance to decide how to debug it.  If none handle this either then open debugger (see UnhandedError-defaultAction)"
	UnhandledError signalForException: self! !

!CMethodToImplement methodsFor: 'printing' stamp: 'Richo 6/20/2011 21:04'!
description
	"Return a textual description of the exception."
	^ 'To implement: ' , methodClass asString , '>>' , methodSelector printString! !

!CMethodToImplement methodsFor: 'accessing' stamp: 'Richo 6/20/2011 21:04'!
methodClass
	^ methodClass! !

!CMethodToImplement methodsFor: 'accessing' stamp: 'Richo 6/20/2011 21:04'!
methodSelector
	^ methodSelector ! !

!CMethodToImplement methodsFor: 'private' stamp: 'Richo 6/20/2011 21:04'!
setClass: aClass selector: aByteSymbol 
	methodClass := aClass.
	methodSelector := aByteSymbol! !

!String methodsFor: '*CTranslator' stamp: 'Richo 6/29/2011 11:18'!
asCIdentifier
"
'Richo' asCIdentifier.
'Richo es un capo' asCIdentifier.
'foo_bar(234)' asCIdentifier
"

^ self inject: '_' into: [:subTotal :next |
	next isAlphaNumeric 
		ifTrue: [subTotal, next asString]
		ifFalse: [subTotal , '_', next asciiValue asString]]! !

Object subclass: #CObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

CObject subclass: #CLinkedList
	instanceVariableNames: 'value next'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CLinkedList class methodsFor: 'definition' stamp: 'Richo 7/1/2011 14:10'!
instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you're doing"
^ #(#CLong #CLinkedList)! !

!CLinkedList class methodsFor: 'c-instance creation' stamp: 'Richo 7/26/2011 16:32'!
with: aNumber
	<argTypes: #(CLong)>
	<returnType: #(instanceSide)>
	^ self new value: aNumber! !

!CLinkedList methodsFor: 'as yet unclassified' stamp: 'Richo 7/1/2011 14:11'!
add: aNumber
	<argTypes: #(CLong)>
	next := CLinkedList new.
	next value: aNumber.! !

!CLinkedList methodsFor: 'as yet unclassified' stamp: 'Richo 7/27/2011 13:36'!
at: aNumber
	<argTypes: #(CLong)>
	<returnType: #(CLong)>
	aNumber = 1
		ifTrue: [^ self value].
	^ next at: aNumber - 1! !

!CLinkedList methodsFor: 'as yet unclassified' stamp: 'Richo 7/27/2011 15:14'!
initialize
	next := 0.
	value := 0! !

!CLinkedList methodsFor: 'as yet unclassified' stamp: 'Richo 7/1/2011 14:10'!
value
	<returnType: #(CLong)>
	^ value! !

!CLinkedList methodsFor: 'as yet unclassified' stamp: 'Richo 7/26/2011 16:32'!
value: aNumber
	<returnType: #(yourself)>
	<argTypes: #(CLong)>
	value := aNumber.
	^ self! !

!CObject class methodsFor: 'converting' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^ self name! !

!CObject class methodsFor: 'c-instance creation' stamp: 'Richo 7/26/2011 15:29'!
basicNew
<returnType: #(instanceSide)>
<creationMethod>
! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:53'!
definition
	"Answer a String that defines the receiver."
	"Overwritten to show the #instanceVariablesTypes: message in the class definition of CObject subclasses"
	| aStream path |
	self = CObject ifTrue: [^super definition].
	aStream := WriteStream
				on: (String new: 300).
	superclass == nil
		ifTrue: [aStream nextPutAll: 'ProtoObject']
		ifFalse: [path := ''.
			self environment
				scopeFor: superclass name
				from: nil
				envtAndPathIfFound: [:envt :remotePath | path := remotePath].
			aStream nextPutAll: path , superclass name].
	aStream nextPutAll: self kindOfSubclass;
		 store: self name.
	[(self hasTraitComposition
			and: [self traitComposition notEmpty])
		ifTrue: [aStream cr; tab; nextPutAll: 'uses: '; nextPutAll: self traitCompositionString]] on: MessageNotUnderstood do: ["Ignore DNU. Traits don't exist in Etoys"].
	aStream cr; tab; nextPutAll: 'instanceVariableNames: '; store: self instanceVariablesString.
	aStream cr; tab; nextPutAll: 'instanceVariableTypes: '; store: self instanceVariableTypes.
	aStream cr; tab; nextPutAll: 'classVariableNames: '; store: self classVariablesString.
	aStream cr; tab; nextPutAll: 'poolDictionaries: '; store: self sharedPoolsString.
	aStream cr; tab; nextPutAll: 'category: '; store: (SystemOrganization categoryOfElement: self name) asString.
	superclass
		ifNil: [aStream nextPutAll: '.';
				 cr.
			aStream nextPutAll: self name.
			aStream space; nextPutAll: 'superclass: nil'].
	^ aStream contents! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:53'!
definitionST80
	"Only useful for the Etoys image"
	self = CObject ifTrue: [^super definitionST80].
	^ self definition! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:53'!
instanceVariableTypes
	"This method will be automaticaly generated for subclasses whenever a subclass of CObject is compiled. It returns the information of the inst vars' type. I choose to do it this way (instead of using an class inst var) to ensure the type info will be saved when we file out this class or store it using Monticello"
	^ #()! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
instanceVariableTypes: anArray
"Automatically compiles a method called #instanceVariableTypes that returns the array given as a parameter"
anArray isEmpty ifTrue: [^ self].
self class compile: 'instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you''re doing"
^ ', anArray asString classified: #definition
! !

!CObject class methodsFor: 'testing' stamp: 'Richo 6/30/2011 12:26'!
isPrimitive
	^ false! !

!CObject class methodsFor: 'c-instance creation' stamp: 'Richo 7/25/2011 15:43'!
new
	<returnType: #(instanceSide)>
	^ self basicNew initialize! !

!CObject class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
subclass: aSymbol category: aString
	^ self subclass: aSymbol instanceVariableNames: '' classVariableNames: '' poolDictionaries: '' category: aString! !

!CObject class methodsFor: 'definition' stamp: 'Richo 7/26/2011 15:24'!
subclass: nameOfClass instanceVariableNames: instVarNames instanceVariableTypes: instVarTypes classVariableNames: classVarNames poolDictionaries: poolDictnames category: category1 
	"Overwritten to allow defining the inst vars' type info of a class"
	^ (self
		subclass: nameOfClass
		instanceVariableNames: instVarNames
		classVariableNames: classVarNames
		poolDictionaries: poolDictnames
		category: category1)
		instanceVariableTypes: instVarTypes ! !

!CObject methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
= t1
<returnType: #(CBoolean)>
<c: '({0} == {1})'>! !

!CObject methodsFor: 'c-pointers' stamp: 'Richo 6/23/2011 16:44'!
address
"	<returnType: #(*)>"
	<c: '&{0}'>! !

!CObject methodsFor: 'c-pointers' stamp: 'Richo 6/23/2011 16:43'!
pointer
	<c: '*{0}'>! !

!CObject methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
~= t1
<returnType: #(CBoolean)>
<c: '({0} !!= {1})'>! !

CObject subclass: #CPrimitiveType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

CPrimitiveType subclass: #CBlock
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CBlock class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	"Blocks are not *really* primitive types, but they're certainly not regular objects. For instance, you can't add inst vars to them. This method should not be implemented and it should'n be called as well."
	self shouldNotImplement! !

!CBlock methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:28'!
doWhileTrue: aBlock
<c: 'do\{{0}\}while({1})'>! !

!CBlock methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
repeat
<c: 'while(true)\{{0}\}'>! !

!CBlock methodsFor: 'c-controlling' stamp: 'Richo 6/21/2011 19:03'!
whileTrue
<c: 'while({0})'>! !

!CBlock methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
whileTrue: aBlock
<c: 'while({0})\{{1}\}'>! !

CPrimitiveType subclass: #CBoolean
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CBoolean class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^'bool'! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
and: aBlock
<c: '({0} && {1})'>! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
ifFalse: falseBlock
<c: 'if(!!({0}))\{{1}\}'>! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
ifFalse: falseBlock ifTrue: trueBlock
<c: 'if({0})\{{2}\}else\{{1}\}'>! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
ifTrue: aBlock
<c: 'if({0})\{{1}\}'>! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
ifTrue: trueBlock ifFalse: falseBlock
<c: 'if({0})\{{1}\}else\{{2}\}'>! !

!CBoolean methodsFor: 'c-logical operations' stamp: 'Richo 6/20/2011 17:29'!
not
<c: '!!({0})'>! !

!CBoolean methodsFor: 'c-controlling' stamp: 'Richo 6/20/2011 17:29'!
or: aBlock
<c: '({0} || {1})'>! !

CPrimitiveType subclass: #CNumber
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

CNumber subclass: #CFloat
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CFloat class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^'float'! !

CNumber subclass: #CLong
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CLong class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^'long'! !

!CNumber methodsFor: 'c-arithmetic' stamp: 'Richo 6/20/2011 17:29'!
* aNumber
<c: '({0} * {1})'>! !

!CNumber methodsFor: 'c-arithmetic' stamp: 'Richo 6/20/2011 17:29'!
+ aNumber
<c: '({0} + {1})'>! !

!CNumber methodsFor: 'c-arithmetic' stamp: 'Richo 6/20/2011 17:29'!
- aNumber
<c: '({0} - {1})'>! !

!CNumber methodsFor: 'c-arithmetic' stamp: 'Richo 6/20/2011 17:29'!
/ aNumber
<c: '({0} / {1})'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
// aNumber
<returnType: #(CLong)> <argTypes: #(CLong)>
^ (self / aNumber) floor! !

!CNumber methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
< aNumber
<returnType: #(CBoolean)>
<c: '({0} < {1})'>! !

!CNumber methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
<= aNumber
<returnType: #(CBoolean)>
<c: '({0} <= {1})'>! !

!CNumber methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
> aNumber
<returnType: #(CBoolean)>
<c: '({0} > {1})'>! !

!CNumber methodsFor: 'c-comparing' stamp: 'Richo 6/22/2011 16:23'!
>= aNumber
<returnType: #(CBoolean)>
<c: '({0} >= {1})'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
\\ aNumber
<returnType: #(CLong)> <argTypes: #(CLong)>
^ self - (self / aNumber * aNumber)! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/20/2011 17:30'!
abs
<c: 'abs({0})'>! !

!CNumber methodsFor: 'c-truncation and round off' stamp: 'Richo 6/20/2011 17:37'!
atRandom
<c: '(Random({0}) + 1)'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
floor
<returnType: #(CLong)> <argTypes: #(CLong)>
^ self! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
isDivisibleBy: aNumber
<returnType: #(CBoolean)> <argTypes: #(CLong)>
aNumber = 0 ifTrue: [^false].
self \\ aNumber = 0
	ifTrue: [^ true]
	ifFalse: [^ false]! !

!CNumber methodsFor: 'c-arithmetic' stamp: 'Richo 6/20/2011 17:30'!
negated
<c: '({0} * -1)'>! !

!CNumber methodsFor: 'c-truncation and round off' stamp: 'Richo 6/20/2011 17:37'!
random
<c: '(Random({0}) + 1)'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
rounded
<returnType: #(CLong)>
self >= 0 
	ifTrue: [^ (self + 0.5) truncated]
	ifFalse: [^ (self - 0.5) truncated]! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/20/2011 17:36'!
sign
<c: 'sign({0})'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
sin
	<returnType: #(CFloat)>
	<tempTypes: #(CFloat CFloat CFloat CFloat)>
	| sum delta self2 i |
	sum := self.
	delta := self.
	self2 := 0 - (self * self).
	i := 2.
	[delta > 0.000000000001]	 whileTrue: [
		"once"
		delta := (delta * self2) / (i * (i + 1)).
		i := i + 2.
		sum := sum + delta.
		"twice"
		delta := (delta * self2) / (i * (i + 1)).
		i := i + 2.
		sum := sum + delta].
	^ sum! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/20/2011 17:36'!
sqrt
<c: 'sqrt({0})'>! !

!CNumber methodsFor: 'c-enumerating' stamp: 'Richo 6/20/2011 17:37'!
timesRepeat: aBlock
<c: 'repeat({0})\{{1}\}'>! !

!CNumber methodsFor: 'c-math functions' stamp: 'Richo 6/22/2011 16:23'!
truncated
<returnType: #(CLong)> <argTypes: #(CLong)>
^ self! !

CNumber subclass: #CUnsignedLong
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CUnsignedLong class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^ 'unsigned long'! !

!CPrimitiveType class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	self subclassResponsibility ! !

!CPrimitiveType class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
definition
	"Answer a String that defines the receiver."
	"Only shows #subclass:category:"
	| aStream path |
	aStream := WriteStream
				on: (String new: 300).
	superclass == nil
		ifTrue: [aStream nextPutAll: 'ProtoObject']
		ifFalse: [path := ''.
			self environment
				scopeFor: superclass name
				from: nil
				envtAndPathIfFound: [:envt :remotePath | path := remotePath].
			aStream nextPutAll: path , superclass name].
	aStream nextPutAll: self kindOfSubclass;
		 store: self name.
	aStream cr; tab; nextPutAll: 'category: '; store: (SystemOrganization categoryOfElement: self name) asString.
	superclass
		ifNil: [aStream nextPutAll: '.';
				 cr.
			aStream nextPutAll: self name.
			aStream space; nextPutAll: 'superclass: nil'].
	^ aStream contents! !

!CPrimitiveType class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
definitionST80
	"Only useful for the Etoys image"
	^ self definition! !

!CPrimitiveType class methodsFor: 'testing' stamp: 'Richo 6/30/2011 11:59'!
isPrimitive
	^ true! !

!CPrimitiveType class methodsFor: 'definition' stamp: 'Richo 6/20/2011 17:27'!
subclass: aSymbol instanceVariableNames: instVars classVariableNames: classVars poolDictionaries: poolDicts category: aString
	"Ignore instance variables, even if the user tries to force the superclass to add them"
	^ super subclass: aSymbol instanceVariableNames: '' classVariableNames: '' poolDictionaries: '' category: aString! !

CPrimitiveType subclass: #CString
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CString class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^'string'! !

!CString methodsFor: 'nxc-concatenation' stamp: 'Richo 6/20/2011 17:37'!
, aNXString
<c: 'StrCat({0}, {1})'>! !

CPrimitiveType subclass: #CVoid
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CVoid class methodsFor: 'as yet unclassified' stamp: 'Richo 6/20/2011 17:52'!
asCType
	^ 'void'! !

CObject subclass: #CProgram
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

CProgram subclass: #CHelloWorld
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CHelloWorld methodsFor: 'as yet unclassified' stamp: 'Richo 6/28/2011 10:47'!
main
	self printf: 'Hello world\n';
		system: ('PAUSE')! !

!CProgram class methodsFor: 'translating' stamp: 'Richo 6/20/2011 21:25'!
translatedToC
	"Useful method to easily translate a program to c. By default the class name is used as program name"
	^ self translatedToC: self name! !

!CProgram class methodsFor: 'translating' stamp: 'Richo 6/20/2011 21:25'!
translatedToC: programName 
	"Useful method to easily translate a program to c"
	^ CProgramTranslator translate: self asProgramNamed: programName! !

!CProgram class methodsFor: 'translating' stamp: 'Richo 6/22/2011 11:12'!
translatedToCpp
	"Useful method to easily translate a program to c++. By default the class name is used as program name"
	^ self translatedToCpp: self name! !

!CProgram class methodsFor: 'translating' stamp: 'Richo 6/22/2011 11:11'!
translatedToCpp: programName 
	"Useful method to easily translate a program to c++"
	^ CppProgramTranslator translate: self asProgramNamed: programName! !

!CProgram methodsFor: 'as yet unclassified' stamp: 'Richo 6/28/2011 10:47'!
printf: aString
	<c: 'printf({1})' >! !

!CProgram methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 16:51'!
printf: aString value: obj
	<c: 'printf({1},{2})' >! !

!CProgram methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 17:54'!
system: str
	<c: 'system({1})'>! !

CProgram subclass: #CProgram1
	instanceVariableNames: 'var1 var2'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram1 class methodsFor: 'definition' stamp: 'Richo 6/22/2011 16:28'!
instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you're doing"
^ #(#CLong #CLong)! !

!CProgram1 methodsFor: 'as yet unclassified' stamp: 'Richo 7/27/2011 15:32'!
main
	| var3 |
	<tempTypes: #(CLong)>
	<returnType: #(CLong)>
	var1 := 3.
	var2 := 4.
	var3 := 17.
	self printf: '%d\n' value: var1 + var2 - var3;
		system: 'PAUSE'.
	^ 42! !

CProgram subclass: #CProgram2
	instanceVariableNames: 'var1 var2'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram2 class methodsFor: 'definition' stamp: 'Richo 6/22/2011 11:33'!
instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you're doing"
^ #(#CLong #CLong)! !

!CProgram2 methodsFor: 'as yet unclassified' stamp: 'Richo 7/27/2011 15:34'!
main
	var2 := 3.
	var1 := 4.
	self printf: '%d\n' value: (self sumar: var2 y: var1);
		system: 'PAUSE'! !

!CProgram2 methodsFor: 'as yet unclassified' stamp: 'Richo 6/22/2011 16:23'!
sumar: a y: b
	<argTypes: #(CLong CLong)>
	<returnType: #(CLong)>
	^ a + b! !

CProgram subclass: #CProgram3
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram3 methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 17:54'!
main
	| p h |
	<tempTypes: #(CLong CLong*)>
	p := 45.
	h := p address.
	self printf: '%d\n' value: h pointer.
	h := h + 1.
	self printf: '%d\n' value: h.
	self system: 'PAUSE'! !

CProgram subclass: #CProgram4
	instanceVariableNames: 'var'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram4 class methodsFor: 'definition' stamp: 'Richo 6/29/2011 17:27'!
instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you're doing"
^ #(#CLong)! !

!CProgram4 methodsFor: 'as yet unclassified' stamp: 'Richo 7/27/2011 17:24'!
main
	| p |
	<tempTypes: #(CProgram4)>
	p := CProgram4 new.
	p var: 5.
	var := 6.
	self print: p; 
		print: self;
		var: 7;
		print: self;
		system: 'PAUSE'! !

!CProgram4 methodsFor: 'as yet unclassified' stamp: 'Richo 7/27/2011 16:36'!
print: p
	<argTypes: #(CProgram4)>
	self printf: '%d\n' value: p var! !

!CProgram4 methodsFor: 'as yet unclassified' stamp: 'Richo 6/29/2011 17:28'!
var
	<returnType: #(CLong)>
	^ var! !

!CProgram4 methodsFor: 'as yet unclassified' stamp: 'Richo 6/29/2011 17:29'!
var: aNumber
	<argTypes: #(CLong)>
	var := aNumber! !

CProgram subclass: #CProgram5
	instanceVariableNames: 'p'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram5 class methodsFor: 'definition' stamp: 'Richo 6/30/2011 13:27'!
instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you're doing"
^ #(#CProgram4)! !

!CProgram5 methodsFor: 'as yet unclassified' stamp: 'Richo 6/30/2011 13:10'!
hello
	self printf: 'Hello World\n'
	! !

!CProgram5 methodsFor: 'as yet unclassified' stamp: 'Richo 7/1/2011 14:07'!
main
	| q |
	<tempTypes: #(CProgram4)>
	self hello.
	p := CProgram4 new.
	p var: 5.
	q := p.
	self printf: '%d\n' value: p var.
	self printf: '%d\n' value: q var.
	p var: 6.
	self printf: '%d\n' value: p var.
	self printf: '%d\n' value: q var.
	self system: 'PAUSE'! !

CProgram subclass: #CProgram6
	instanceVariableNames: 'p var'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram6 class methodsFor: 'definition' stamp: 'Richo 7/1/2011 14:03'!
instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you're doing"
^ #(#CProgram6 #CLong)! !

!CProgram6 methodsFor: 'as yet unclassified' stamp: 'Richo 7/1/2011 14:04'!
hello
	self printf: 'Hello World: %d\n' value: self var! !

!CProgram6 methodsFor: 'as yet unclassified' stamp: 'Richo 7/1/2011 14:04'!
main
	p := CProgram6 new.
	p var: 3.
	p hello.
	var := 4.
	self hello.
	self system: 'PAUSE'! !

!CProgram6 methodsFor: 'as yet unclassified' stamp: 'Richo 7/1/2011 14:04'!
var
	<returnType: #(CLong)>
	^ var! !

!CProgram6 methodsFor: 'as yet unclassified' stamp: 'Richo 7/1/2011 14:04'!
var: aNumber
	<argTypes: #(CLong)>
	var := aNumber! !

CProgram subclass: #CProgram7
	instanceVariableNames: 'list'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram7 class methodsFor: 'definition' stamp: 'Richo 7/1/2011 14:10'!
instanceVariableTypes
"This method was autogenerated. Do not change unless you *really* know what you're doing"
^ #(#CLinkedList)! !

!CProgram7 methodsFor: 'as yet unclassified' stamp: 'Richo 7/14/2011 16:28'!
main
	list := CLinkedList new.
	list value: 1;
		add: 2.
	self print: (list at: 1);
		print: (list at: 2);
		pause ! !

!CProgram7 methodsFor: 'as yet unclassified' stamp: 'Richo 7/14/2011 17:00'!
pause
	<inlined>
	self system: 'PAUSE'! !

!CProgram7 methodsFor: 'as yet unclassified' stamp: 'Richo 7/14/2011 17:30'!
print: n
	<inlined>
	<argTypes: #(CLong)>
	self printf: '%d\n' value: n.
! !

CProgram subclass: #CProgram8
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgram8 methodsFor: 'as yet unclassified' stamp: 'Richo 7/26/2011 16:35'!
main
	| list |
	<tempTypes: #(CLinkedList)>
	list := CLinkedList with: 1.
	list add: 2.
	self printf: '%d\n' value: list value.
	self printf: '%d\n' value: (list at: 2).
	self pause.! !

!CProgram8 methodsFor: 'as yet unclassified' stamp: 'Richo 7/15/2011 13:54'!
pause
	<inlined>
	self system: 'PAUSE'! !

CProgram subclass: #CProgramInline2Test
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgramInline2Test methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 19:36'!
main
	self prueba: 'Richo'.
	^ self prueba * 4! !

!CProgramInline2Test methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 19:36'!
prueba
	| var |
	<inlined>
	<tempTypes: #(CLong)>
	<returnType: #(CLong)>
	var := 42.
	^ 57! !

!CProgramInline2Test methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 19:37'!
prueba: arg
	<inlined>
	<argTypes: #(CString)>
	<returnType: #(CString)>
	^ arg! !

CProgram subclass: #CProgramInlineTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Programs'!

!CProgramInlineTest methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 13:27'!
main
	| var1 var2 |
	<tempTypes: #(CLong CUnsignedLong)>
	var2 := 1.
	var1 := var2 * 3.
	self pause.
	self prueba: var1 * var2.
	var1 := 5.
	self prueba: var2 * var1.
	^ 5! !

!CProgramInlineTest methodsFor: 'as yet unclassified' stamp: 'Richo 7/21/2011 13:06'!
pause
	<inlined>
	self system: 'PAUSE'! !

!CProgramInlineTest methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 13:30'!
prueba2
	<inlined>
	<returnType: #(CLong)>
	^ 5! !

!CProgramInlineTest methodsFor: 'as yet unclassified' stamp: 'Richo 7/22/2011 13:30'!
prueba: arg
	<inlined>
	<argTypes: #(CLong)>
	^ self prueba2 * 2! !

Object subclass: #COutputBuilder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!COutputBuilder commentStamp: 'Richo 6/28/2011 12:03' prior: 0!
A COutputBuilder is responsible for constructing a set of objects that represent the output from the translators.!

!COutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:31'!
functionClass
	^ CTranslatedFunction! !

!COutputBuilder class methodsFor: 'building' stamp: 'Richo 7/8/2011 18:07'!
newFunctionFor: ownerClass node: methodNode body: bodySource arguments: argVars temporaries: tempVars
	^ self functionClass class: ownerClass node: methodNode body: bodySource arguments: argVars temporaries: tempVars! !

!COutputBuilder class methodsFor: 'building' stamp: 'Richo 6/28/2011 11:24'!
newProgramNamed: aByteSymbol 
	^ self programClass named: aByteSymbol! !

!COutputBuilder class methodsFor: 'building' stamp: 'Richo 6/28/2011 11:25'!
newStructFor: aClass 
	^ self structClass for: aClass! !

!COutputBuilder class methodsFor: 'building' stamp: 'Richo 6/28/2011 11:49'!
newVariableNamed: aByteString type: aClass 
	^ self variableClass name: aByteString type: aClass! !

!COutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:24'!
programClass
	^ CTranslatedProgram! !

!COutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 16:05'!
structClass
	^ CTranslatedStruct! !

!COutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 16:06'!
variableClass
	^ CTranslatedVariable! !

COutputBuilder subclass: #CppOutputBuilder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CppOutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:59'!
functionClass
	^ CppTranslatedMethod! !

!CppOutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 11:59'!
programClass
	^ CppTranslatedProgram! !

!CppOutputBuilder class methodsFor: 'factory' stamp: 'Richo 6/28/2011 16:05'!
structClass
	^ CppTranslatedClass! !

Object subclass: #CPointer
	instanceVariableNames: 'type'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Hierarchy'!

!CPointer class methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 17:10'!
on: aClass 
	^ (self basicNew setType: aClass) initialize! !

!CPointer methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 17:21'!
asCType
	^ type asCType , '*'! !

!CPointer methodsFor: 'private' stamp: 'Richo 6/23/2011 17:11'!
setType: aClass 
	type := aClass! !

!CPointer methodsFor: 'as yet unclassified' stamp: 'Richo 6/23/2011 17:21'!
ultimateCompiledMethodAt: aByteSymbol ifAbsent: aBlockClosure 
	^ type ultimateCompiledMethodAt: aByteSymbol ifAbsent: aBlockClosure ! !

Object subclass: #CPrimitiveRule
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CPrimitiveRule commentStamp: 'Richo 6/28/2011 10:53' prior: 0!
A CPrimitiveRule is is a very simple way of defining a rule to translate a smalltalk message to a c primitive. Every function (such as "printf" or "system"), every operator (such as "+" or "=="), and every control structure (such as "if", "while", or "repeat") is considered to be a c primitive. To define a method as a c primitive you should use the #c: pragma (look at the senders of #c: for examples).

To create a rule simply send '#as: CPrimitiveRule' to a well formed string. For example:

	'NumOut({2}, {3}, {1})' as: CPrimitiveRule.
	'({0} + {1})' as: CPrimitiveRule.
	'if({0})\{{1}\}else\{{2}\}' as: CPrimitiveRule.
	
Once a rule is created it can be applied to a message node using a translator by sending #applyTo:using:. This mehod will then replace all '{n}' in the string with the translated arguments ({0} represents the receiver).

Instance Variables
	value:		<String>

value
	- The string that defines how to apply the rule.
!

!CPrimitiveRule class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 20:59'!
newFrom: aString
	^ self on: aString! !

!CPrimitiveRule class methodsFor: 'instance creation' stamp: 'Richo 6/28/2011 12:06'!
on: aByteString 
	^ self basicNew initializeWith: aByteString! !

!CPrimitiveRule methodsFor: 'applying' stamp: 'Richo 6/20/2011 21:13'!
applyTo: aMessageNode using: aCMethodTranslator 
	"Visits the message node receiver and arguments and replaces the '{n}' in the string with the result ({0} represents the receiver). It reinitializes the source stream for every node so that we only get the result of visiting that node"
	^ self format: ({aMessageNode receiver}, aMessageNode arguments
		collectWithIndex: [:node :index | 
		(String streamContents: [:stream | 
				aCMethodTranslator withSource: stream do: [aCMethodTranslator visitNode: node].
				self removeSemicolonIfNecessaryOn: stream for: index - 1])])! !

!CPrimitiveRule methodsFor: 'applying' stamp: 'Richo 6/20/2011 20:59'!
format: aCollection
	"Zero based. It goes through value increasing by 1 the enclosed expressions and then it sends regular #format:"
	| oldStream newStream |
	oldStream := value readStream.
	newStream := WriteStream on: ''.
	[oldStream atEnd]
		whileFalse: [| next |
		next := oldStream next.
		newStream nextPut: next.
		(next = ${ and: [oldStream peekBack ~= $\])
			ifTrue: [| n | "We found an enclosed expression"
			n := oldStream upTo: $}.
			newStream nextPutAll: (n asNumber + 1) asString, '}'.
			]
	].
	^ newStream contents format: aCollection
	! !

!CPrimitiveRule methodsFor: 'initializing' stamp: 'Richo 6/28/2011 12:07'!
initializeWith: aByteString 
	self initialize.
	value := aByteString ! !

!CPrimitiveRule methodsFor: 'printing' stamp: 'Richo 6/20/2011 20:59'!
printOn: aStream
super printOn: aStream.
aStream nextPut: $(.
aStream nextPutAll: value.
aStream nextPut: $)! !

!CPrimitiveRule methodsFor: 'applying' stamp: 'Richo 6/20/2011 20:59'!
removeSemicolonIfNecessaryOn: source for: argIndex
	"Depending on the position of the argument in the rule string, it can be necessary to remove the semicolon after the statement. For now, we only leave the semicolon if the token is surrounded by curly brackets {}.
	For instance, in the rule below
	
	'while({0})\{{1}\}'
	
	The node replacing the {0} token will have its semicolon removed, whereas the {1} won't. Of course, most of the time it won't be necessary to remove any semicolon (if {0} is a literal or a variable, for instance)."
	| index newContents oldContents rule token |
	oldContents := source contents.
	rule := value withBlanksCondensed.
	
	"Do nothing if the source does not end with a semicolon"
	(oldContents size > 2 and: [oldContents endsWith: ';', String cr]) ifFalse: [^self].

	token := '{', argIndex asString ,'}'.
	index := rule findString: token.

	"Do nothing if the token is not in the rule"
	index = 0 ifTrue: [^self].

	"Do nothing if the token is not surrounded by curly brackets"
	((rule at: index - 2 ifAbsent: ['']) asString ,
	(rule at: index - 1 ifAbsent: ['']) asString ,
	(rule at: index + token size ifAbsent: ['']) asString ,
	(rule at: index + token size + 1 ifAbsent: ['']) asString) = '\{\}'
		ifTrue: [^self].

	"Okay, you can now safely remove the semicolon"
	newContents := oldContents allButLast: 2.
	source reset; nextPutAll: newContents.! !

Object subclass: #CProgramTranslator
	instanceVariableNames: 'program'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CProgramTranslator commentStamp: 'Richo 6/28/2011 10:51' prior: 0!
A CProgramTranslator is responsible for translating subclasses of CProgram from smalltalk to C. It's public protocol is very simple, just use #translate:asProgramNamed:, which returns an instance of CTranslatedProgram.
An even easier way of translating programs is sending #translatedToC to a CProgram subclass. For example:

	CHelloWorld translatedToC.
	
You can also specify a different program name:

	CHelloWorld translatedToC: 'Other program name'.

The translation is done recursively, starting from the #main method and translating every method call found from there. For more information, see CMethodTranslator comment.


Instance Variables
	program:		<CTranslatedProgram>

program
	- The program being translated.
!

!CProgramTranslator class methodsFor: 'output builder' stamp: 'Richo 6/28/2011 11:23'!
builder
	^ COutputBuilder! !

!CProgramTranslator class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 20:59'!
new
	self shouldNotImplement! !

!CProgramTranslator class methodsFor: 'instance creation' stamp: 'Richo 6/28/2011 11:33'!
translate: aCProgram asProgramNamed: programName
	^ (self basicNew initializeWithProgram: (self builder newProgramNamed: programName))
			translate: aCProgram! !

!CProgramTranslator methodsFor: 'output builder' stamp: 'Richo 6/28/2011 11:25'!
builder
	^ self class builder! !

!CProgramTranslator methodsFor: 'initializing' stamp: 'Richo 6/28/2011 11:34'!
initializeWithProgram: aCTranslatedProgram 
	program := aCTranslatedProgram.
	self initialize! !

!CProgramTranslator methodsFor: 'testing' stamp: 'Richo 7/1/2011 16:37'!
isAlreadyTranslated: selector on: class
	^ program methods anySatisfy: [:each | 
		each name = selector asCIdentifier and: [each belongsTo: class]]! !

!CProgramTranslator methodsFor: 'accessing' stamp: 'Richo 6/21/2011 15:33'!
methodTranslator
	"Override if you want to define a specific method translator"
	^ CMethodTranslator! !

!CProgramTranslator methodsFor: 'translating' stamp: 'Richo 6/28/2011 11:25'!
translate: aCProgram
	program addStruct: (self builder newStructFor: aCProgram).
	[self translateMethod: #main of: aCProgram]
		on: CInstVarToImplement do: [:ex |
			(program addStruct: (self builder newStructFor: ex ownerClass))
				addVar: (self builder newVariableNamed: ex varName type: ex varType).
			ex resume].
	^ program! !

!CProgramTranslator methodsFor: 'translating' stamp: 'Richo 7/1/2011 16:36'!
translateMethod: selector of: class
	"Recursively translates this method and all methods called from it"
	[program addMethod: (self methodTranslator translate: selector of: class with: self builder)]
		on: CMethodToImplement
		do: [:exception | 
			((selector = exception methodSelector and: [class = exception methodClass])
				or: [self isAlreadyTranslated: exception methodSelector on: exception methodClass])
				ifFalse: [self translateMethod: exception methodSelector of: exception methodClass].
			exception resume]! !

CProgramTranslator subclass: #CppProgramTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Translators'!

!CppProgramTranslator class methodsFor: 'output builder' stamp: 'Richo 6/28/2011 12:01'!
builder
	^ CppOutputBuilder ! !

!CppProgramTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 6/22/2011 11:09'!
methodTranslator
	"Override if you want to define a specific method translator"
	^ CppMethodTranslator! !

!CppProgramTranslator methodsFor: 'as yet unclassified' stamp: 'Richo 6/28/2011 11:28'!
translateMethod: selector of: class
	"Recursively translates this method and all methods called from it"
	[program
		addStruct: (self builder newStructFor: class); "Just to ensure we have a struct for this class"
		addMethod: (self methodTranslator translate: selector of: class with: self builder)]
		on: CMethodToImplement
		do: [:exception | 
			self translateMethod: exception methodSelector of: exception methodClass.
			exception resume]! !

Object subclass: #CTranslatedFunction
	instanceVariableNames: 'methodNode ownerClass body arguments temporaries'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CTranslatedFunction class methodsFor: 'instance creation' stamp: 'Richo 7/8/2011 18:08'!
class: ownerClass node: methodNode body: bodySource arguments: argVars temporaries: tempVars
	^ self basicNew initializeWithClass: ownerClass node: methodNode body: bodySource arguments: argVars temporaries: tempVars! !

!CTranslatedFunction methodsFor: 'comparing' stamp: 'Richo 6/29/2011 11:29'!
= anObject
	(anObject isKindOf: self class) ifFalse: [^ false].
	^ self source = anObject source! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 7/8/2011 20:18'!
arguments
	^ arguments ! !

!CTranslatedFunction methodsFor: 'testing' stamp: 'Richo 7/27/2011 16:06'!
belongsTo: aClass
	^ ownerClass instanceSide = aClass instanceSide! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 7/8/2011 20:22'!
declaration
	^ String
		streamContents: [:stream | 
			| returnType |
			returnType := self returnType.
			returnType isPrimitive ifFalse: [returnType := CPointer on: returnType].
			stream nextPutAll: returnType asCType.
			stream nextPutAll: ' ' , self name , '('.
			self arguments
				do: [:each | stream nextPutAll: each source]
				separatedBy: [stream nextPutAll: ', '].
			stream nextPutAll: ');']! !

!CTranslatedFunction methodsFor: 'comparing' stamp: 'Richo 6/20/2011 20:49'!
hash
	^ self source hash hashMultiply ! !

!CTranslatedFunction methodsFor: 'initializing' stamp: 'Richo 7/8/2011 18:09'!
initializeWithClass: aClass node: aMethodNode body: aByteString arguments: argVars temporaries: tempVars
	ownerClass := aClass.
	methodNode := aMethodNode.
	body := aByteString.
	arguments := argVars.
	temporaries := tempVars.
	self initialize.! !

!CTranslatedFunction methodsFor: 'testing' stamp: 'Richo 6/30/2011 12:18'!
isMain
	^ self name = '_main'! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 6/30/2011 12:17'!
name
	^ methodNode selector asCIdentifier ! !

!CTranslatedFunction methodsFor: 'printing' stamp: 'Richo 6/30/2011 12:03'!
printOn: aStream
	aStream nextPutAll: self source! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 6/30/2011 12:02'!
returnType
	^ CTypeCollector findTypeOf: methodNode in: ownerClass! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 6/30/2011 12:02'!
selfType
	^ ownerClass! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 7/8/2011 20:27'!
source
	^ String streamContents: [:stream |
		stream nextPutAll: self declaration allButLast; cr.
		stream nextPut: ${; cr.
		self temporaries do: [:each | stream nextPutAll: each source, ';'; cr].
		stream nextPutAll: body.
		stream nextPut: $}
		]! !

!CTranslatedFunction methodsFor: 'accessing' stamp: 'Richo 7/8/2011 20:18'!
temporaries 
	^ temporaries ! !

CTranslatedFunction subclass: #CppTranslatedMethod
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CppTranslatedMethod methodsFor: 'accessing' stamp: 'Richo 7/27/2011 16:28'!
declaration
	^ String streamContents: [:stream || returnType |
		returnType := self returnType.
		returnType isPrimitive ifFalse: [returnType := CPointer on: returnType].
		ownerClass isMeta 
			ifTrue: [stream nextPutAll: 'static '].
		stream nextPutAll: returnType asCType, ' '.
		"ownerClass isMeta 
			ifFalse: [stream nextPutAll: ownerClass name, '::']."
		stream nextPutAll: self name, '('.
		self arguments
				do: [:each | stream nextPutAll: each source]
				separatedBy: [stream nextPutAll: ', '].
		stream nextPutAll: ');']! !

Object subclass: #CTranslatedProgram
	instanceVariableNames: 'structs methods programName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CTranslatedProgram class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 20:49'!
named: programName
	^ (self basicNew setName: programName) initialize! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/29/2011 11:32'!
addMethod: aCTranslatedMethod
	methods add: aCTranslatedMethod ! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 7/27/2011 16:01'!
addStruct: aCStruct
	"Adds a struct for the given class only if a struct is not already added. Returns the struct"
	^ structs
		detect: [:each | each ownerClass instanceSide = aCStruct ownerClass instanceSide]
		ifNone: [structs add: aCStruct]! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/30/2011 12:12'!
header
	"Store program's name"
	| source |
	source := WriteStream on: ''.
	source nextPutAll: '/* ', self headerFileName ,' */'; cr; cr.

	"Store structs definitions"
	structs do: [:struct |
		source nextPutAll: struct source; cr.
	].
	source cr.

	"Store methods' declarations"
	source nextPutAll: self mainDeclaration; cr.
	methods do: [:method |
			source nextPutAll: method declaration; cr.
		].
	source cr.

	^ source contents! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 10:51'!
headerFileName
	^ programName, '.h'! !

!CTranslatedProgram methodsFor: 'initialize-release' stamp: 'Richo 6/29/2011 11:31'!
initialize
	super initialize.
	structs := Set new.
	methods := Set new! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/30/2011 12:13'!
mainDeclaration
	^ (self mainImplementation upTo: Character cr) , ';'! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 7/1/2011 13:53'!
mainImplementation
	^'int main()
\{
{1}* program = ({1}*)malloc(sizeof({1}));
{2}(program);
return 0;
\}' format: {
	self mainMethod selfType asCType.
	self mainMethod name}! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/29/2011 11:23'!
mainMethod
	^ methods detect: [:each | each isMain]! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
methods
	^ methods! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
name
	^ programName ! !

!CTranslatedProgram methodsFor: 'printing' stamp: 'Richo 6/21/2011 17:21'!
printOn: aStream
	aStream
		nextPutAll: self header; cr;
		nextPutAll: '=================='; cr;
		nextPutAll: self source! !

!CTranslatedProgram methodsFor: 'fileOut' stamp: 'Richo 6/21/2011 17:36'!
saveInDirectory: aDirectory
	aDirectory
		forceNewFileNamed: self headerFileName do: [:file | file nextPutAll: self header];
		forceNewFileNamed: self sourceFileName do: [:file | file nextPutAll: self source]! !

!CTranslatedProgram methodsFor: 'private' stamp: 'Richo 6/28/2011 10:51'!
setName: aString
	programName := aString asCIdentifier! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/29/2011 11:31'!
source
	"Store program's name"
	| source |
	source := WriteStream on: ''.
	source nextPutAll: '/* ', self sourceFileName ,' */'; cr; cr.

	"Store header include"
	source nextPutAll: '#include "', self headerFileName, '"'; cr; cr.
	
	"Store methods implementation"
	source nextPutAll: self mainImplementation; cr; cr.
	methods do: [:method |
			source nextPutAll: method source]
			separatedBy: [source cr; cr].

	^ source contents! !

!CTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 10:51'!
sourceFileName
	^ programName , '.c'! !

CTranslatedProgram subclass: #CppTranslatedProgram
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CppTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 7/27/2011 16:06'!
addMethod: aCppTranslatedMethod
	(structs detect: [:each | aCppTranslatedMethod belongsTo: each ownerClass])
		addMethod: aCppTranslatedMethod.
	methods add: aCppTranslatedMethod! !

!CppTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/22/2011 11:58'!
header
	"Store program's name"
	| source |
	source := WriteStream on: ''.
	source nextPutAll: '/* ', self headerFileName ,' */'; cr; cr.

	"Store structs definitions"
	structs do: [:struct |
		source nextPutAll: struct source; cr.
	].
	source cr.
	
	^ source contents! !

!CppTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 7/27/2011 15:20'!
mainImplementation
	^'int main()
\{
{1}* program = new {1}();
program->{2}();
return 0;
\}' format: {
	self mainMethod selfType asCType.
	self mainMethod name}

! !

!CppTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 7/27/2011 16:31'!
source
	"Store program's name"
	| source |
	source := WriteStream on: ''.
	source nextPutAll: '/* ', self sourceFileName ,' */'; cr; cr.

	"Store header include"
	source nextPutAll: '#include "', self headerFileName, '"'; cr; cr.
	
	"Store main method implementation. The methods are included in the header inside their respective classes"
	source nextPutAll: self mainImplementation; cr; cr.

	^ source contents! !

!CppTranslatedProgram methodsFor: 'accessing' stamp: 'Richo 6/28/2011 10:51'!
sourceFileName 
	^ programName, '.cpp'! !

Object subclass: #CTranslatedStruct
	instanceVariableNames: 'ownerClass variables'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CTranslatedStruct class methodsFor: 'instance creation' stamp: 'Richo 6/28/2011 12:07'!
for: aClass 
	^ self basicNew initializeWith: aClass! !

!CTranslatedStruct methodsFor: 'comparing' stamp: 'Richo 6/28/2011 11:55'!
= anObject
	(anObject isKindOf: self class) ifFalse: [^ false].
	^ anObject ownerClass = ownerClass and: [anObject variables = variables]! !

!CTranslatedStruct methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:50'!
addVar: aCInstVar 
	variables add: aCInstVar ! !

!CTranslatedStruct methodsFor: 'comparing' stamp: 'Richo 6/28/2011 11:52'!
hash
	^ ownerClass hash hashMultiply bitXor: variables hash! !

!CTranslatedStruct methodsFor: 'initialize-release' stamp: 'Richo 6/28/2011 12:09'!
initializeWith: aClass
	self initialize.
	ownerClass := aClass.
	variables := Set new.! !

!CTranslatedStruct methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
ownerClass
	^ ownerClass ! !

!CTranslatedStruct methodsFor: 'accessing' stamp: 'Richo 7/27/2011 16:45'!
source
	^ String streamContents: [:source || className |
		className := ownerClass asCType.
		"Class side"
		source nextPutAll: 'struct _', className , '_class {'; cr.
		source nextPutAll: '};'; cr.
		source nextPutAll: 'typedef struct _', className, '_class ', className, '_class;'; cr.
		source nextPutAll: className , '_class The_', className, '_class;'; cr; cr.
		"Instance side"
		source nextPutAll: 'struct _', className, ' {'; cr.
		variables do: [:var |
			var type = ownerClass
				ifTrue: [source nextPutAll: 'struct _'].
			source nextPutAll: var source, ';'; cr].
		source nextPutAll: '};'; cr.
		source nextPutAll: 'typedef struct _', className, ' ', className, ';'		
		]! !

!CTranslatedStruct methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
variables
	^ variables ! !

CTranslatedStruct subclass: #CppTranslatedClass
	instanceVariableNames: 'methods'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CppTranslatedClass methodsFor: 'accessing' stamp: 'Richo 6/22/2011 11:48'!
addMethod: aCppTranslatedMethod 
	methods add: aCppTranslatedMethod.! !

!CppTranslatedClass methodsFor: 'initialize-release' stamp: 'Richo 6/22/2011 11:48'!
initialize
	super initialize.
	methods := Set new.! !

!CppTranslatedClass methodsFor: 'accessing' stamp: 'Richo 7/27/2011 16:42'!
source
	^ String streamContents: [:source |
		source nextPutAll: 'class ', ownerClass asCType, ' {'; cr.
		source nextPutAll: 'private:'; cr; cr.
		variables do: [:var |
			source nextPutAll: var source; nextPut: $;; cr].
		source cr; nextPutAll: 'public:'; cr; cr.
		methods do: [:method |
			source nextPutAll: method source; cr]
			separatedBy: [source cr].
		source nextPutAll: '};']! !

Object subclass: #CTranslatedVariable
	instanceVariableNames: 'varName varType'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CTranslator-Output'!

!CTranslatedVariable class methodsFor: 'instance creation' stamp: 'Richo 6/20/2011 20:49'!
name: varName type: varType
	^ (self basicNew setName: varName type: varType) initialize! !

!CTranslatedVariable methodsFor: 'comparing' stamp: 'Richo 6/20/2011 20:51'!
= anObject
(anObject isKindOf: self class) ifFalse: [^ false].
^ self name = anObject name and: [self type = anObject type]! !

!CTranslatedVariable methodsFor: 'comparing' stamp: 'Richo 6/20/2011 20:49'!
hash
^ (self name hash hashMultiply bitXor: self type hash) hashMultiply ! !

!CTranslatedVariable methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
name
^ varName ! !

!CTranslatedVariable methodsFor: 'private' stamp: 'Richo 6/20/2011 20:50'!
setName: aString type: aClass
	varName := aString.
	varType := aClass ! !

!CTranslatedVariable methodsFor: 'accessing' stamp: 'Richo 7/25/2011 15:39'!
source
	| type |
	self type isMeta ifTrue: [^ self type instanceSide asCType , '_class ', self name].
	type := self type isPrimitive ifTrue: [self type] ifFalse: [CPointer on: self type].
	^ type asCType, ' ', self name! !

!CTranslatedVariable methodsFor: 'accessing' stamp: 'Richo 6/20/2011 20:49'!
type
^ varType! !
